/*--------------- COPYRIGHT ------------------------
| INRA - Unit MIA de Jouy en Josas                |
--------------------------------------------------*/

/*--------------- IDENTIFICATION PRODUIT -----------
| Derniere mise a jour : 28 aout 2008              |
| Role                 : gre la mthode par       |
|                        grilles                   |
--------------------------------------------------*/
////////////////////////////////////////////
#include "caliinclc.h"
#include "mt19937ar.h"
#ifdef RAND48
#include <stdlib.h>		// pour srand48
#endif
#include "methodGrid.h"
#include "calimacros.h"
#include "geom.h"
#include "intersection.h"
#include "timing.h"
#include "zoneintegration.h"



  ///////////////////////////////////////////////
  // CONSTRUCTORS
  ///////////////////////////////////////////////

methodGrid::methodGrid ():methodIntegr ()
{
}

////////////////////////////////////////////////////

methodGrid::methodGrid (const int nfunc, const int *ifunct):
methodIntegr (nfunc, ifunct)
{
}

////////////////////////////////////////////////////
methodGrid::methodGrid (const int nfunc,
			const int *ifunct,
			const unsigned int seed,
			const real l, const real h, const int est):
methodIntegr (nfunc, ifunct)
{
  int ifunc, i;
  this->l = l;
  this->h = h;
  this->est = est;
  this->pseed = seed;
  this->nbeval = 0;

  for (ifunc = 0; ifunc < this->nfunct; ifunc++)
    {
      i = ifunct[ifunc] - 1;
    }				// fin ifunc


  /* Allocation des structures pour les sommes de M */
  CREER_T1 (this->sommeM, (MAX_TRIANGLES * MAX_TRIANGLES), tPolygoni);
  CREER_T1 (this->k, (MAX_TRIANGLES * MAX_TRIANGLES), int);

}				// fin constructor


///////////////////////////////////////////////////////
// Destructeur
///////////////////////////////////////////////////////

methodGrid::~methodGrid ()
{
  // Ne rien dsallouer  ce niveau

}

/////////////////////////////////////////////
// Verifie les arguments de la mthode
/////////////////////////////////////////////
int
methodGrid::VerifArgu ()
{
  char moi[] = "methodGrid::VerifArgu";
/* Pour stocker les messages */
  char errmess[CHAR_MAX];

  int code = OK;
  int ifunc, i;

  if ((code = this->VerifFunct ()) != OK)
    {
      return (code);
    }

  if (this->l <= 0)
    {
      sprintf (errmess, "Invalid step: must be positive.\n");
      code = CALI_ERGRID1;
      ecrmess (CALI_ERGRID1, moi, errmess);
    }
  if (this->h <= 0)
    {
      sprintf (errmess, "Invalid step: must be positive.\n");
      code = CALI_ERGRID1;
      ecrmess (CALI_ERGRID1, moi, errmess);
    }

  if ((this->est < 2) || (this->est > MAX_EST))
    {
      sprintf (errmess,
	       "Invalid number of estimations: must be greater or equal to 2 and less or equal to %d\n",
	       MAX_EST);
      code = CALI_ERGRID2;
      ecrmess (CALI_ERGRID2, moi, errmess);
    }


  for (ifunc = 0; ifunc < this->nfunct; ifunc++)
    {
      i = ifunct[ifunc] - 1;
    }


  return (code);
}				// fin methodGrid::VerifArgu


/////////////////////////////////////////////
  // Read the method arguments
  ///////////////////////////////////////////////


int
methodGrid::ReadArgu ()
{
  char moi[] = "methodGrid::ReadArgu";
/* Pour stocker les messages */
  char errmess[CHAR_MAX]; 
  char errlect[]="Error in reading";
  int i;

  printf ("Step for integration x axis (in meter):  ");
  if (scanf ("%lf", &(this->l))<1) {
    return (ecrmess (CALI_ERLECT, moi, errlect, True));
  }
  this->l = this->l * SCALE;

  printf ("Step for integration y axis (in meter):  ");
  if (scanf ("%lf", &(this->h))<1) {
    return (ecrmess (CALI_ERLECT, moi, errlect, True));
  }
  this->h = this->h * SCALE;

  this->est = 1;
  while ((this->est < 2) || (this->est > MAX_EST))
    {
      printf ("Number of estimations? ([2-%d]) ", MAX_EST);
      if (scanf ("%d", &(this->est))<1) 
	return (ecrmess (CALI_ERLECT, moi, errlect, True));

      if ((this->est < 2) || (this->est > MAX_EST))
	{
	  sprintf (errmess,
		   "Invalid number of estimations: must be greater than 2 and less or equal to %d\n",
		   MAX_EST);
	  return (ecrmess (CALI_ERGRID2, moi, errmess));
	}
    }


  printf ("\nSeed of the random generator: ");
  if (scanf ("%d", &i)<1)
    return (ecrmess (CALI_ERLECT, moi, errlect, True));
// Rque: lire directement dans seed => warning de compil car seed est unsigned
  this->pseed = int (i);

  return (this->VerifArgu ());


}				// fin ReadArgu

/////////////////////////////////////////////
  // Print the results
  ///////////////////////////////////////////////

void
methodGrid::Print (const int poutput, const real areac, const real aread)
{
  real coefp[MAX_NFUNCTIONS];
  real areacc, areadd;
  int i, ifunc;
  real Tol = REAL_MIN;		// Pour la comparaison des rls

  areacc = areac / (SCALE * SCALE);
  areadd = aread / (SCALE * SCALE);



  if ((areacc <= 0) || (areadd <= 0))
    {
      printf ("\n Careful:\n");
      if (areacc <= 0)
	printf ("   area of polygon 1 is nul\n");
      if (areadd <= 0)
	printf ("   area of polygon 2 is nul\n");
    }
  else
    {
      for (ifunc = 0; ifunc < this->nfunct; ifunc++)
	{
	  if (fabs (this->rp[ifunc]) <= Tol)	// if (this->rp[ifunc] == 0.0
	    coefp[ifunc] = 0.0;
	  else
	    coefp[ifunc] = (this->ep[ifunc] / this->rp[ifunc]);

	  if ((poutput != NOTHING) && (poutput != LIGHT))
	    {
	      printf ("\nIntegrated flow for function %d:\n",
		      this->ifunct[ifunc]);
	      printf (" mean: %g mean/area1: %g mean/area2: %g\n",
		      this->rp[ifunc], (this->rp[ifunc] / areacc),
		      (this->rp[ifunc] / areadd));
	      fflush (stdout);
	    }
	  if (this->methcalcul[ifunc] == True)
	    {
	      if ((poutput == ALL) && (this->nrepet[ifunc] > 0))
		{
		  printf
		    (" standard deviation: %g \n coefficient of variation (std/mean): %g\n",
		     this->ep[ifunc], coefp[ifunc]);
		  fflush (stdout);
		}
	      i = ifunct[ifunc] - 1;

	    }			// fin methcalcul
	}			// fin nfunc

      if (poutput == ALL)
	printf ("\narea1: %g area2: %g \n", areacc, areadd);
      else
	printf ("\n");
      fflush (stdout);

    }				// fin else

}				// fin Print

//////////////////////////////////////////////////////////////
// Print a line on the result-file
//////////////////////////////////////////////////////////////
void
methodGrid::PrintFic (FILE * fp, const int noa, const int nob,
		      const long int temps,
		      const real areac, const real aread)
{

  int ifunc, nr;
  // Ecrire les infos communes aux 2 mthodes:
  this->PrintFicOut (fp, noa, nob, areac, aread);
  if (OUTPUT_FILE_FORMAT == ALL)
    fprintf (fp, "\t%ld", temps);
  fprintf (fp, "\n");
  fflush (fp);
  if (noa != nob)
    {
      this->PrintFicOut (fp, nob, noa, aread, areac);
      if (OUTPUT_FILE_FORMAT == ALL)
	fprintf (fp, "\t%ld", temps);
      fprintf (fp, "\n");
      fflush (fp);
    }				// fin noa

  // Ecrire ce qui est spcifique  'grid':
  if ((OUTPUT_FILE_FORMAT == ALL) || (OUTPUT_FILE_FORMAT == LIGHT))
    {
      fprintf (fp, "%g\t%g", (this->l / SCALE), (this->h / SCALE));
      fprintf (fp, "\t%ld\n", this->nbeval);
      fflush (fp);

      for (ifunc = 0; ifunc < this->nfunct; ifunc++)
	{
	  fprintf (fp, "%g\t%g", this->rp[ifunc], this->ep[ifunc]);
	  fprintf (fp, "\t%d\tr:", this->nrepet[ifunc]);
	  for (nr = 0; nr < this->nrepet[ifunc]; nr++)
	    {
	      fprintf (fp, "\t%g", this->resultp[ifunc][nr]);
	    }
	  fprintf (fp, "\n");
	  fflush (fp);
	}			// fin nfunc
    }				// fin format ALL ou LIGHT

}				// fin PrintFic


//////////////////////////////////////////////////////////////
// Integration Version grille:
/////////////////////////////////////////////////////////
void
methodGrid::Integration (Function * dispf,
			 tPolygoni A, tPolygoni B,
			 int aii, int bii, real h, real l,
			 tPolygoni sommeM, int k, int nr, long int &nbptseval)
{

  int i, j, c, ifunc;
  int ixmax, ixmin, iymax, iymin;
  Boolean w, z;
  real sa, area, areascale;
  real rx, ry, x, y;		// valeurs alatoires

  tPointd T1, t;
  tPolygond D;
  tdVertex intersection;
  tdVertex pointer, ipointer;

  /* Les resultats pour le couple de sous-polys */
  real rst[MAX_NFUNCTIONS];

  char moi[] = "methodGrid::Integration";
/* Pour stocker les messages */
  char errmess[CHAR_MAX];



  for (ifunc = 0; ifunc < this->nfunct; ifunc++)
    rst[ifunc] = 0;

  i = 0;


  /* Calcul de la grille */

  ixmax = sommeM[i][XX];
  ixmin = sommeM[i][XX];
  iymax = sommeM[i][YY];
  iymin = sommeM[i][YY];

  do
    {
      i = i + 1;

      if (sommeM[i][XX] > ixmax)
	ixmax = sommeM[i][XX];
      else if (sommeM[i][XX] < ixmin)
	ixmin = sommeM[i][XX];
      if (sommeM[i][YY] > iymax)
	iymax = sommeM[i][YY];
      else if (sommeM[i][YY] < iymin)
	iymin = sommeM[i][YY];
    }
  while (i < k - 1);


  sa = h * l;

/* r is a random floating point value in the range [0,1) {including 0, not including 1}. Note we must convert rand() and/or RAND_MAX+1 to floating point values to avoid integer division. In addition, Sean Scanlon pointed out the possibility that RAND_MAX may be the largest positive integer the architecture can represent, so (RAND_MAX+1) may result in an overflow, or more likely the value will end up being the largest negative integer the architecture can represent, so to avoid this we convert RAND_MAX and 1 to doubles before adding. */

// AB: juin 2007: un nbre aleatoire different sur l'axe des x et des y
#ifdef RAND48
  rx = (real) drand48 ();
  ry = (real) drand48 ();
#else
#ifdef RAND
  rx = ((real) rand () / ((real) (RAND_MAX) + (real) (1)));
  ry = ((real) rand () / ((real) (RAND_MAX) + (real) (1)));
#else
  rx = genrand_real2 ();
  ry = genrand_real2 ();
#endif
#endif


  x = (rx * l);
  y = (ry * h);




  j = 0;
  while ((t[YY] = iymin + j * h + y) <= iymax)
    {
      i = 0;
      while ((t[XX] = ixmin + i * l + x) <= ixmax)
	{

	  nbptseval++;

	  w = InPolyConvex (t, sommeM, k);



	  if (w)
	    {

	      for (c = 0; c < bii; c++)
		{
		  Assd (T1, B[c]);
		  SubVecd (T1, t, D[c]);

		}
	      // D= B-t

	      NEW (intersection, tdsVertex);
	      intersection->next = intersection->prev = intersection;
	      //  Initialisation du 1ier sommet
	      intersection->v[XX] = intersection->v[YY] = 0;
	      //AB: 18/06/2007
	      // Regarder si un des poly de l'intersection est inclus

	      // dans l'autre car ConvexIntersect ne dtecte pas 
	      // ce cas



	      z = ConvexInclus (A, aii, D, bii, intersection);




	      if (z == False)
		{
		  // Si ce n'est pas le cas:
		  // calculer l'intersection eventuelle

		  z = ConvexIntersect (A, aii, D, bii, intersection);
		}

	      if (z == True)
		{

		  // AB: verif rajoutee en juin 2007
		  if (intersection->next->next == intersection)
		    {
		      sprintf (errmess,
			       "Intersection with 2 points only 1: %g %g\n 2: %g %g\n",
			       intersection->v[XX], intersection->v[YY],
			       intersection->next->v[XX],
			       intersection->next->v[YY]);
// Fatal error
		      ecrmess (CALI_ERINTERNAL, moi, errmess, True);
		    }


		  if ((intersection->next == intersection) ||
		      (intersection->prev == intersection))
		    {
		      sprintf (errmess,
			       "Intersection with 1 point only 1: %g %g\n",
			       intersection->v[XX], intersection->v[YY]);
// Fatal error
		      ecrmess (CALI_ERINTERNAL, moi, errmess, True);
		    }

		  if ((intersection->v[XX] == 0.0) ||
		      (intersection->v[YY] == 0.0))

		    {
		      sprintf (errmess, "Intersection with no point.\n");

// Fatal error
		      ecrmess (CALI_ERINTERNAL, moi, errmess, True);
		    }

		  /* Fin des verif rajoutees */


		  area = polygon_area_2 (intersection) / 2;



		  // Remarque: pas besoin de dsallouer  quand z faux
		  // i.e quand l'intersection est vide
		  pointer = intersection->next;
		  do
		    {
		      ipointer = pointer;
		      pointer = ipointer->next;
		      FREE (ipointer);
		    }
		  while (pointer != intersection);

		}
	      else
		area = 0;

	      // Calcul du produit des aires en m2
	      // Diviser par SCALE**4 car produit d'aires
	      areascale = area / (SCALE * SCALE * SCALE * SCALE);

	      FREE (intersection);

	      //Creation du point en entre de la fonction de dispersion
	      Point mont (t[XX], t[YY]);
	      for (ifunc = 0; ifunc < this->nfunct; ifunc++)
		{
		  if (this->methcalcul[ifunc] == True)
		    rst[ifunc] +=
		      ((*dispf[this->ifunct[ifunc] - 1]) (mont) * areascale);
		}		// fin ifunc

	    }
	  i++;
	}
      j++;
    }




  for (ifunc = 0; ifunc < this->nfunct; ifunc++)
    {
      if (this->methcalcul[ifunc] == True)
	{
	  rst[ifunc] = rst[ifunc] * sa;
	  this->resultp[ifunc][nr] += rst[ifunc];

	}			// fin this->methcalcul[ifunc]
    }				// fin ifunc


}				// fin Integration version grille





/////////////////////////////////////////////
  // Estimation
  ///////////////////////////////////////////////
void
methodGrid::CalcR (const int poutput,
		   Function * dispf,
		   const real areac, const real aread,
		   const real mindist, const Point lepoint,
		   const int numbera, const int numberb,
		   const int ac, const int ad,
		   const int nic[MAX_VERTICES],
		   const int nid[MAX_VERTICES],
		   tPolygoni * Polyc, tPolygoni * Polyd, long &temps)
{
  // Pour les heuristiques:
  int nintegre = this->nfunct;

  int i, ino, ifunc, j, itr, nr;

  // Mesure du temps
  long int oldTime;
  oldTime = timing ();

  long int nbptseval = 0;	// nbre de points d'eval


  nintegre = this->nfunct;

  // c'est le mme pour toutes les fonctions

  for (ifunc = 0; ifunc < this->nfunct; ifunc++)
    {
      this->nrepet[ifunc] = 0;
      this->rp[ifunc] = 0.0;
      this->ep[ifunc] = 0.0;
      this->methcalcul[ifunc] = True;
      for (nr = 0; nr < this->est; nr++)
	this->resultp[ifunc][nr] = 0.0;

      /* Ne faire l'intgration que si necessaire */

      /* La fonction s'annule-t-elle? */
      if ((this->dzero[this->ifunct[ifunc] - 1] > 0.0) &&
	  (mindist >= this->dzero[this->ifunct[ifunc] - 1]))
	{
	  /* Les points les plus proches sont loigns de plus
	     de la distance  partir de laquelle la fonction 
	     s'annule */
	  if (poutput == ALL)
	    printf
	      ("Minimal distance between polygons %d,%d=%g (>=%g):\n   function %d set to zero.\n",
	       numbera, numberb,
	       mindist,
	       this->dzero[this->ifunct[ifunc] - 1], this->ifunct[ifunc]);
	  nintegre--;
	  this->methcalcul[ifunc] = False;
	  this->rp[ifunc] = 0;
	}
      else
	{
	  /* Calcule-t-on   entre les barycentres? */
	  if ((this->dpoint[this->ifunct[ifunc] - 1] > 0.0) &&
	      (mindist >= this->dpoint[this->ifunct[ifunc] - 1]))
	    {
	      /* on calcule entre les barycentres */

	      if (poutput == ALL)
		printf
		  ("Minimal distance between polygons %d,%d=%g (>=%g):\n   fonction %d calculated between centroids.\n",
		   numbera, numberb,
		   mindist, this->dpoint[this->ifunct[ifunc] - 1],
		   this->ifunct[ifunc]);
	      /* Les points les plus proches sont loigns de plus
	         de la distance  partir de laquelle on ne calcule
	         la fonction qu'en un point */
	      nintegre--;
	      this->methcalcul[ifunc] = False;
	      this->rp[ifunc] =
		((areac / (SCALE * SCALE)) *
		 (aread / (SCALE * SCALE)) *
		 (*dispf[this->ifunct[ifunc] - 1]) (lepoint));
	    }			// fin (mindist >= dpoint[ifunc] )

	}
    }				// fin ifunc

  if (nintegre > 0)
    {
      /*  Il reste des fonctions a integrer */
      /* Boucle sur les sous-polygones: calcul de la somme de M */
      itr = 0;
      for (i = 0; i < ac; i++)	// i: indice de sous-polys dans Polyc
	for (j = 0; j < ad; j++)
	  {

	    this->k[itr] = SommeMinkowski (Polyc[i], nic[i],
					   Polyd[j], nid[j],
					   this->sommeM[itr]);



	    itr++;
	  }			// fin i,j

      if (poutput == ALL)
	printf ("\nIntegrated flows at each replication:");

      /* Initialisation de la graine du gnrateur aleatoire:
         on l'initialise pour chaque couple de poly */
#ifdef RAND48
      srand48 (this->pseed);
#else
#ifdef  RAND
      srand (this->pseed);
#else
      init_genrand (this->pseed);
#endif
#endif



      for (nr = 0; nr < this->est; nr++)
	{

	  /* Boucle sur les sous-polygones */
	  itr = 0;
	  for (i = 0; i < ac; i++)	// i: indice de triangles dans Polyc
	    for (j = 0; j < ad; j++)
	      {
		this->Integration (dispf,
				   Polyc[i], Polyd[j], nic[i], nid[j],
				   this->h, this->l, this->sommeM[itr],
				   this->k[itr], nr, nbptseval);
		// Les rsultats  sont dans les resultp
		itr++;
	      }			// fin boucle j,i


	  /* Boucle sur les fonctions */
	  for (ifunc = 0; ifunc < this->nfunct; ifunc++)
	    {
	      ino = ifunct[ifunc] - 1;


	      this->nrepet[ifunc]++;	/* nbre de repet effectives */
	      if (this->methcalcul[ifunc] == True)
		{
		  if (poutput == ALL)
		    {
		      printf ("\n%d. Function %d: %g ",
			      (nr + 1),
			      this->ifunct[ifunc], this->resultp[ifunc][nr]);
		    }
		  // pour le calcul des moyennes
		  this->rp[ifunc] += this->resultp[ifunc][nr];


		}		// fin this->methcalcul
	    }			// fin ifunc


	}			// fin boucle sur les repet (nr)


      if (poutput == ALL)
	printf ("\n");

      for (ifunc = 0; ifunc < this->nfunct; ifunc++)
	{
	  if (this->methcalcul[ifunc] == True)
	    {
	      this->rp[ifunc] = this->rp[ifunc] / (this->nrepet[ifunc]);
	      // this->nrepet est le nbre de repet effectivement ralises
	      // contrairement  this->est qui est celui demand
	      // Calcul de l'ecart-type (standard dev.)
	      for (i = 0; i < this->nrepet[ifunc]; i++)
		{
		  this->ep[ifunc] +=
		    pow (this->resultp[ifunc][i] - this->rp[ifunc], 2);
		}

	      // Remarque: "nrepet" est  toujours > 1
	      this->ep[ifunc] =
		sqrt (this->ep[ifunc] / (this->nrepet[ifunc] - 1));
	    }			// fin (this->methcalcul[ifunc] == False)

	}			// fin ifunc
    }				// fin nintegre>0

  temps = timing () - oldTime;
  this->nbeval = nbptseval;

  if (poutput == ALL)
    {
      printf ("\nElapsed time in integration: %ld milliseconds\n", temps);

      printf ("Nb. evaluations: %ld\n", nbptseval);
      fflush (stdout);

    }


}				//  fin CalcR

///////////////////////////////////////////
