/*--------------- COPYRIGHT ------------------------
| INRA - Unit MIA de Jouy en Josas                |
--------------------------------------------------*/

/*--------------- IDENTIFICATION PRODUIT -----------
| Derniere mise a jour : 3 Mai 2006                |
| Role                 : suppression de sommets    |
| aligns de polygones lus sur un fichier          |
--------------------------------------------------*/
///////////////////////////////////////////////////////
#include <iostream>

using std::cerr;
using std::endl;
#include "caliinclc.h"
#include "caliconfig.h"
#include "calierror.h"
#include "calimacros.h"
#include "readPoly.h"
#include "geom.h"
#include "util.h"
 //////////////////////////////////////////////////////////
// Liberer la memoire
//////////////////////////////////////////////////////////
void
libMem (int npoly,
	tPolygond ** Polyd,
	int *a, int *numPoly, real * area,
	int **ni, char **nomPoly, tPolygoni ** Poly, real ** bary)
{
  DETRU_T2 (Polyd, npoly);
  libMemPoly (npoly, a, numPoly, area, ni, nomPoly, Poly, bary);
}				// fin libMem

///////////////////////////////////////////////////////
// Ecriture sur fichier  des polygones corrigs
//  format 1
///////////////////////////////////////////////////////
void
ecr1Poly (FILE * fp, char *pdelim, int ID, int nsom, tPolygond Poly)
{
  int i;
  fprintf (fp, "%d%s", ID, pdelim);
  for (i = 0; i < nsom; i++)
    fprintf (fp, "%.3f%s", Poly[i][XX], pdelim);
  fprintf (fp, "\n%d%s", ID, pdelim);
  for (i = 0; i < nsom; i++)
    fprintf (fp, "%.3f%s", Poly[i][YY], pdelim);
  fprintf (fp, "\n");

}

///////////////////////////////////////////////////////
// Ecriture sur fichier  des polygones corrigs
//  format 2
///////////////////////////////////////////////////////
void
ecr2Poly (FILE * fp, char *pdelim, int ID, char *nom, int nsom,
	  tPolygond Poly)
{
  int i;
  fprintf (fp, "%d%s%s%s%d\n", ID, pdelim, nom, pdelim, nsom);
  for (i = 0; i < nsom; i++)
    fprintf (fp, "%.3f%s", Poly[i][XX], pdelim);
  fprintf (fp, "\n");
  for (i = 0; i < nsom; i++)
    fprintf (fp, "%.3f%s", Poly[i][YY], pdelim);
  fprintf (fp, "\n");
}

///////////////////////////////////////////////////////
int
polysimplif (char *filenamei, char *filenamer, Boolean pverbose)
{

  // On ne le demande pas les paramtres suivants:
  // on leur donne les valeurs par dfaut:
  int pinput = DEFAULT_INPUT_FORMAT;
  char pdelim[] = DEFAULT_DELIM;
//////////////////////////////////////////////////////////

  int npoly;

// Va de travail
  int i, erreur;
  char lu[MAX_LINE_POLY];
  FILE *fres = NULL, *fpi = NULL;

  char moi[] = "polysimplif";
/* Pour stocker les messages */
  char errmess[CHAR_MAX];
  char errlect[]="Error in reading";
//////////////////////////////////////////////////////////
  // Lecture du fichier de polygones
//////////////////////////////////////////////////////////

  fpi = fopen (filenamei, "r");
  if (!fpi)
    {
      sprintf (errmess, "cannot open polygons file %s\n", filenamei);
      return (ecrmess (CALI_ERFIC1, moi, errmess));
    }

  // Lire avec fgets ce fichier, car il a un nbre va. de donnes
  // sur chaque ligne et il ne faut pas mlanger les fscanf et fgets
  if (fgets (lu, MAX_LINE_POLY, fpi) == NULL)
    return (ecrmess (CALI_ERLECT, moi, errlect, True));
  npoly = atoi (lu);
  //  printf ("\nNumber of polygons: %d\n-------------------\n", npoly);


//////////////////////////////////////////////////////////
  // Declaration des structures de dim npoly:
//////////////////////////////////////////////////////////
  int *a;
  int **ni;
  tPolygoni **Poly;
  tPolygond **Polyd;
  int *numPoly;			/* les nos d'identification des polys */
  char **nomPoly;		/* les noms des polys */
  real **bary;			/* Les barycentres des polys (inutilis) */

//////////////////////////////////////////////////////////
  // Allocation des structures de dim npoly:
// memory allocation of the structures of dimension=npoly
///////////////////////////////////////////////////////
  CREER_T1 (a, npoly, int);
  CREER_T1 (numPoly, npoly, int);
  CREER_T2 (ni, npoly, int);
  CREER_T2 (nomPoly, npoly, char);
  CREER_T2 (Poly, npoly, tPolygoni);
  CREER_T2 (bary, npoly, real);
  CREER_T2 (Polyd, npoly, tPolygond);

  // Les tailles =1: on ne dcoupe pas en sous-polys
  for (i = 0; i < npoly; i++)
    {
      CREER_T1 (ni[i], 1, int);
      CREER_T1 (nomPoly[i], MAX_NAME, char);
      CREER_T1 (Poly[i], 1, tPolygoni);
      CREER_T1 (Polyd[i], 1, tPolygond);
      CREER_T1 (bary[i], DIM, real);
    }

//////////////////////////////////////////////////////////
// Lecture et correction des polygones
//////////////////////////////////////////////////////////

  real *area = NULL;		// pas besoin des aires
  int npolybons;
  erreur = ReadPoly (fpi, pverbose, False, pinput, pdelim,
		     npoly, nomPoly, numPoly, npolybons,
		     a, area, ni, Poly, Polyd, bary);
  if ((erreur != OK) && (erreur != CALI_WARNPOLY))
    {
      libMem (npoly, Polyd, a, numPoly, area, ni, nomPoly, Poly, bary);
      return (erreur);		// fatal error
    }


  if (npolybons == 0)
    {
      libMem (npoly, Polyd, a, numPoly, area, ni, nomPoly, Poly, bary);
      sprintf (errmess,
	       "All polygons are invalid; file %s not created\n", filenamer);
      return (ecrmess (CALI_ERPOLY7, moi, errmess));
    }


//////////////////////////////////////////////////////////
// crire le fichier corrig
//////////////////////////////////////////////////////////

  fres = fopen (filenamer, "w");
  if (!fres)
    {
      libMem (npoly, Polyd, a, numPoly, area, ni, nomPoly, Poly, bary);
      sprintf (errmess, "cannot open result file %s\n", filenamer);
      return (ecrmess (CALI_ERFIC1, moi, errmess));
    }

  // Ecrire le nbre de polys rsultats
  fprintf (fres, "%d\n", npolybons);


  for (i = 0; i < npoly; i++)
    {
      if (numPoly[i] > 0)
	{
	  // n'ecrire que les poly valides
	  switch (pinput)
	    {
	    case 1:
	      ecr1Poly (fres, pdelim, numPoly[i], ni[i][0], Polyd[i][0]);
	      break;
	    case 2:
	      ecr2Poly (fres, pdelim, numPoly[i], nomPoly[i], ni[i][0],
			Polyd[i][0]);
	      break;
	    default:
	      libMem (npoly, Polyd, a, numPoly, area, ni, nomPoly, Poly,
		      bary);
	      sprintf (errmess, "Internal error:  format %d non reconnu\n",
		       pinput);
	      return (ecrmess (CALI_ERINTERNAL, moi, errmess, True));
	      // c'est une erreur de programmation fatale si on est ici
	    }			// fin switch
	}			// fin (numPoly[i]>0)
    }				// fin i

  printf ("\nCreated file: %s\n", filenamer);
  ///////////////////////////////////////
  // Librer la mmoire
  ///////////////////////////////////////
  libMem (npoly, Polyd, a, numPoly, area, ni, nomPoly, Poly, bary);

  ///////////////////////////////////////
  // Fermer les fichiers
  ///////////////////////////////////////
  // Pas besoin de fermer fpi: readPoly l'a fait
  if (fres != NULL)
    fclose (fres);

  ///////////////////////////////////////
  // Retour
  ///////////////////////////////////////
  return OK;
}				// fin polysimplif

///////////////////////////////////////////////////////
// Pour crire un message en cas de mauvaise utilisation 
// de la commande ou help
///////////////////////////////////////////////////////
int
errmain (char c, char *cde)
{
/* Pour stocker les messages */
  char errmess[CHAR_MAX];
  char bad[]="bad use";

  if (c != 'h')
    {
      // cas d'erreur:
      sprintf (errmess,
	       "Bad option %c\nUsage:\n %s -h\nor\n %s -i poly-filename  -r result-filename\n",
	       c, cde, cde);
      return (ecrmess (CALI_ERSYNTAXE, bad, errmess));
    }
  else
    {
      // cas o on veut l'aide
      printf ("Usage: %s -i poly-filename  -r result-filename\n", cde);
      return (OK);
    }
}				// fin errmain

///////////////////////////////////////////////////////
// Programme principal
///////////////////////////////////////////////////////
int
main (int argc, char *argv[])
{
  char *pfilenamei, *pfilenamer;
  char c;
  int iarg;
  int narg = argc - 1;

  // True si on veut l'ip des coordonnes lues
  // et de l'indication si les poly sont convexes ou non  
  Boolean pverbose = Boolean (DEFAULT_VERBOSE);

  // Vrifier le nombre d'arguments
  if (narg != 4)
    return (errmain ('h', argv[0]));

// Init du fichier rsultat
  pfilenamer = NULL;

  // Dcoder les arguments:
  iarg = 1;
  while (iarg <= narg)
    {
      if (argv[iarg][0] == '-')
	{
	  c = argv[iarg][1];
	  switch (c)
	    {
	    case 'h':
	      return (errmain (c, argv[0]));
	    case 'v':
	      pverbose = True;
	      iarg = iarg - 1;
	      break;
	    case 'i':
	      pfilenamei = argv[iarg + 1];
	      break;
	    case 'r':
	      pfilenamer = argv[iarg + 1];
	      break;
	    default:
	      return (errmain (c, argv[0]));
	      break;
	    }			// fin switch (c)
	  iarg = iarg + 2;
	}			// fin if (argv[iarg][0]== '-')
      else
	{
	  return (errmain (c, argv[0]));
	}
    }				// fin while


  return (polysimplif (pfilenamei, pfilenamer, pverbose));

}				// fin main
