/*--------------- COPYRIGHT ------------------------
| INRA - Unit MIA de Jouy en Josas                |
--------------------------------------------------*/

/*--------------- IDENTIFICATION PRODUIT -----------
| Derniere mise a jour : 15 Fev 2006               |
| Role                 : Calcul de la surface      |
|         de polygones lus sur un fichier          |
--------------------------------------------------*/
///////////////////////////////////////////////////////
#include <iostream>

using std::cerr;
using std::endl;

#include "caliinclc.h"
#include "caliconfig.h"
#include "calierror.h"
#include "calimacros.h"
#include "readPoly.h"
#include "util.h"


///////////////////////////////////////////////////////
///////////////////////////////////////////////////////
int
surface (char *filenamei, char *filenamer, Boolean pverbose)
{

  // On ne le demande pas les paramtres suivants:
  // on leur donne les valeurs par dfaut:
  int pinput = DEFAULT_INPUT_FORMAT;
  char pdelim[] = DEFAULT_DELIM;
//////////////////////////////////////////////////////////

  // Les polygones:
  int npoly;



// Va de travail
  int i, erreur;
  char lu[MAX_LINE_POLY];
  FILE *fres = NULL, *fpi = NULL;

  char moi[] = "surface";
/* Pour stocker les messages */
  char errmess[CHAR_MAX];
  char errlect[]="Error in reading";

//////////////////////////////////////////////////////////
  // Lecture du fichier de polygones
//////////////////////////////////////////////////////////

  fpi = fopen (filenamei, "r");
  if (!fpi)
    {
      sprintf (errmess, "cannot open polygons file %s\n", filenamei);
      return (ecrmess (CALI_ERFIC1, moi, errmess));
    }

  // Lire avec fgets ce fichier, car il a un nbre va. de donnes
  // sur chaque ligne et il ne faut pas mlanger les fscanf et fgets
  if (fgets (lu, MAX_LINE_POLY, fpi)==NULL)
    return (ecrmess (CALI_ERLECT, moi, errlect, True));
  npoly = atoi (lu);
  //  printf ("\nNumber of polygons: %d\n-------------------\n", npoly);


//////////////////////////////////////////////////////////
  // Declaration des structures de dim npoly:
//////////////////////////////////////////////////////////
  int *a;
  real *area;
  int **ni;
  tPolygoni **Poly;
  int *numPoly;			/* les nos d'identification des polys */
  char **nomPoly;		/* les noms des polys */
  real **bary;			/* Les barycentres des polys (inutilis) */

//////////////////////////////////////////////////////////
  // Allocation des structures de dim npoly:
// memory allocation of the structures of dimension=npoly
///////////////////////////////////////////////////////
  CREER_T1 (a, npoly, int);
  CREER_T1 (numPoly, npoly, int);
  CREER_T1 (area, npoly, real);
  CREER_T2 (ni, npoly, int);
  CREER_T2 (nomPoly, npoly, char);
  CREER_T2 (Poly, npoly, tPolygoni);
  CREER_T2 (bary, npoly, real);

  for (i = 0; i < npoly; i++)
    {
      CREER_T1 (ni[i], MAX_TRIANGLES, int);
      CREER_T1 (nomPoly[i], MAX_NAME, char);
      CREER_T1 (Poly[i], MAX_TRIANGLES, tPolygoni);
      CREER_T1 (bary[i], DIM, real);
    }




//////////////////////////////////////////////////////////
// Lecture des polygones
//////////////////////////////////////////////////////////
  int npolybons;
  Boolean calcSurf = True;	// on ne veut que les surfaces
  tPolygond **Polyd = NULL;	// on ne veut pas les polys corrigs
  erreur = ReadPoly (fpi, pverbose, calcSurf, pinput, pdelim,
		     npoly, nomPoly, numPoly, npolybons,
		     a, area, ni, Poly, Polyd, bary);

  if ((erreur != OK) && (erreur != CALI_WARNPOLY))
    {
      libMemPoly (npoly, a, numPoly, area, ni, nomPoly, Poly, bary);
      return (erreur);		// fatal error
    }

  if (npolybons == 0)
    {
      libMemPoly (npoly, a, numPoly, area, ni, nomPoly, Poly, bary);
      sprintf (errmess, "All polygons are invalid\n");
      return (ecrmess (CALI_ERPOLY7, moi, errmess));
    }

//////////////////////////////////////////////////////////
// crire les surfaces
//////////////////////////////////////////////////////////

  for (i = 0; i < npoly; i++)
    {
      // Revenir au mtre:
      area[i] = area[i] / (SCALE * SCALE);
    }

  if (filenamer != NULL)
    {

      fres = fopen (filenamer, "w");
      if (!fres)
	{
	  libMemPoly (npoly, a, numPoly, area, ni, nomPoly, Poly, bary);

	  sprintf (errmess, "cannot open file %s\n", filenamer);
	  return (ecrmess (CALI_ERFIC1, moi, errmess));
	}

      // Ecrire le nbre de polys et le nom du fichier d'entree
      fprintf (fres, "npoly:\t%d\tinput-file:\t%s\n", npolybons, filenamei);


      for (i = 0; i < npoly; i++)
	{
	  if (numPoly[i] > 0)
	    {
	      fprintf (fres, "%d\t%g\n", numPoly[i], area[i]);
	    }
	}			// fin i

      printf ("\nCreated file: %s\n", filenamer);
    }				// fin filenamer
  else
    {
      for (i = 0; i < npoly; i++)
	{
	  if (numPoly[i] > 0)
	    {
	      // crire sur stdout
	      printf ("Ident:\t%d", numPoly[i]);
	      if (nomPoly[i] != "")
		printf ("\tname:\t%s", nomPoly[i]);
	      printf ("\tarea:\t%g\n", area[i]);
	    }
	}			// fin i
    }				// fin else


  ///////////////////////////////////////
  // Librer la mmoire
  ///////////////////////////////////////

  libMemPoly (npoly, a, numPoly, area, ni, nomPoly, Poly, bary);

  ///////////////////////////////////////
  // Fermer les fichiers
  ///////////////////////////////////////
  // Pas besoin de fermer fpi: readPoly l'a fait
  if (fres != NULL)
    fclose (fres);

  ///////////////////////////////////////
  // Retour
  ///////////////////////////////////////

  return OK;
}				// fin surface

///////////////////////////////////////////////////////
// Pour crire un message en cas de mauvaise utilisation 
// de la commande ou help
///////////////////////////////////////////////////////
int
errmain (char c, char *cde)
{
/* Pour stocker les messages */
  char errmess[CHAR_MAX];
  char bad[]="bad use";

  if (c != 'h')
    {
      // cas d'erreur:
      sprintf (errmess,
	       "Bad option %c\nUsage:\n %s -h\nor\n %s -i poly-filename  [-r result-filename]\n",
	       c, cde, cde);
      return (ecrmess (CALI_ERSYNTAXE, bad, errmess));
    }
  else
    {
      // cas o on veut l'aide
      printf ("Usage: %s -i poly-filename  [-r result-filename]\n", cde);
      return (OK);
    }
}				// fin errmain

///////////////////////////////////////////////////////
// Programme principal
///////////////////////////////////////////////////////
int
main (int argc, char *argv[])
{
  char *pfilenamei, *pfilenamer;
  char c;
  int iarg;
  int narg = argc - 1;

  // True si on veut l'ip des coordonnes lues
  // et de l'indication si les poly sont convexes ou non  
  Boolean pverbose = Boolean (DEFAULT_VERBOSE);

  // Vrifier le nombre d'arguments
  if ((narg < 1) || (narg > 5))
    return (errmain ('h', argv[0]));

// Init du fichier rsultat
  pfilenamer = NULL;

  // Dcoder les arguments:
  iarg = 1;
  while (iarg <= narg)
    {
      if (argv[iarg][0] == '-')
	{
	  c = argv[iarg][1];
	  switch (c)
	    {
	    case 'h':
	      return (errmain (c, argv[0]));
	    case 'v':
	      pverbose = True;
	      iarg = iarg - 1;
	      break;
	    case 'i':
	      pfilenamei = argv[iarg + 1];
	      break;
	    case 'r':
	      pfilenamer = argv[iarg + 1];
	      break;
	    default:
	      return (errmain (c, argv[0]));
	      break;
	    }			// fin switch (c)
	  iarg = iarg + 2;
	}			// fin if (argv[iarg][0]== '-')
      else
	{
	  return (errmain (c, argv[0]));
	}
    }				// fin while


  return (surface (pfilenamei, pfilenamer, pverbose));

}				// fin main
