/*--------------- COPYRIGHT ------------------------
| INRA - Unite MIA de Jouy en Josas                |
--------------------------------------------------*/

///////////////////////////////////////////////////////

#include <stdio.h>
#include <limits.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include <stdlib.h>

////////////////////////////////////////////////////
// Les fonctions de l'API R
////////////////////////////////////////////////////
extern char * REprintf( char *, ...);
extern char * Rprintf(char *, ...);


////////////////////////////////////////////////////
// Les fonctions geometriques
////////////////////////////////////////////////////

////////////////////////////////////////////////////
// Les includes
////////////////////////////////////////////////////
#include "lesdefines.h"
#include "lestypes.h"
#include "lesmacros.h"

// Note: en R les lignes d'une meme colonne sont stockees 
// consecutivement en memoire.
#define Adp(a, b) ( Ad[nsommets * (b) + (a)])

/////////////////////////////////////////////////
tVertex
MakeNullVertex (tVertex vertices)
{
  tVertex v;
  // AB: il faut eliminer le 0,0 car la decomposition en sous-poly
  // consiste a des triangles qui passent par l'origine
  // mais ceci impose qu'il n'y a pas de poly qui passe par 0,0
  if (vertices->v[XX] == 0 && vertices->v[YY] == 0)
    return vertices;
  else
    {
      NEW (v, tsVertex);

      ADD (vertices, v);

      return v;
    }
}

/////////////////////////////////////////////////////////////////
// Fonction de comparaison de reels  a tolerance pres
/////////////////////////////////////////////////////////////////
int
realequal (double a, double b)
{
  if (fabs (a - b) < (double)  DBL_MIN)
    return 1;
  else
    return 0;
}


////////////////////////////////////////////////////
// Area for a triangle
real
Area2 (tPointd a, tPointd b, tPointd c)
{
  real ax, ay, bx, by, cx, cy, aire;
  ax = (real) a[XX];
  ay = (real) a[YY];
  bx = (real) b[XX];
  by = (real) b[YY];
  cx = (real) c[XX];
  cy = (real) c[YY];
  aire = ((bx - ax) * (cy - ay) - (cx - ax) * (by - ay));

  return (aire);

}
//////////////////////////////////////////////////////
int
LeftOnd (tPointd a, tPointd b, tPointd c)
{
  //  return (int (Area2 (a, b, c) >= real (0.0)));
  return( (Area2 (a, b, c) >= (real) 0.0) ? (1)  : (0));

}
//////////////////////////////////////////////////////


int
LeftOn (tPointi a, tPointi b, tPointi c)
{
  // return (int (Area2i (a, b, c) >= 0));

  real aire;
  tPointd aa, bb, cc;
  aa[XX] = (real) a[XX];
  aa[YY] = (real) a[YY];
  bb[XX] = (real) b[XX];
  bb[YY] = (real) b[YY];
  cc[XX] = (real) c[XX];
  cc[YY] = (real) c[YY];

  aire = Area2 (aa, bb, cc);
  if (aire >= 0.0)
      return(1);
  else
	  return(0);

}
///////////////////////////////////////////////////////


int
Convexity (tVertex vertices)
{
  tVertex v0, v1, v2;
  v1 = vertices;

  do
    {
      v2 = v1->next;
      v0 = v1->prev;

      if (LeftOn (v0->v, v2->v, v1->v))
	return (0);


      v1 = v1->next;
    }
  while (v1 != vertices);
  return (1);
}


//////////////////////////////////////////////////////
// Fonctions pour Triangulate
//////////////////////////////////////////////////////////
int
Left (tPointi a, tPointi b, tPointi c)
{
  // Renvoie 1 si a est a gche de (b,c)
  // AB:3/2/2006
  // ceci peut provoquer un depassement de la capacite entiers
  //  return (int (Area2i (a, b, c) > 0));
  tPointd aa, bb, cc;
  aa[XX] = (real) a[XX];
  aa[YY] = (real) a[YY];
  bb[XX] = (real) b[XX];
  bb[YY] = (real) b[YY];
  cc[XX] = (real) c[XX];
  cc[YY] = (real) c[YY];

  if ( (int)Area2 (aa, bb, cc) >0 )
    return(1);
  else
    return(0);

  //  return (int (Area2 (aa, bb, cc) > real (0.0)));

}

///////////////////////////////////////////////////


int
InCone (tVertex a, tVertex b)
{
  tVertex a0, a1;		// a0,a,a1 are consecutive vertices

  a1 = a->next;
  a0 = a->prev;

  // If a is a convex vertex ...
  if (LeftOn (a->v, a1->v, a0->v) == 1)
    return ((int) (Left (a->v, b->v, a0->v) && Left (b->v, a->v, a1->v)));

  // Else a is reflex:
  return ((int)
	  (!(LeftOn (a->v, b->v, a1->v) && LeftOn (b->v, a->v, a0->v))));
}
/////////////////////////////////////////
void
Assd (tPointd b, tPointi a)
{
  int i;
  for (i = 0; i < DIM; i++)
    b[i] = (real)a[i];

}

////////////////////////////////////////
int
AreaSigni (tPointi a, tPointi b, tPointi c)
{
  real area2;
  tPointd aa,  bb,  cc;
  Assd(aa,a);
  Assd(bb,b);
  Assd(cc,c);

  area2 = Area2 (aa, bb, cc);


    /* The area should be an integer. */
    if      ( area2 >  0.5 ) return  1;
    else if ( area2 < -0.5 ) return -1;
    else                     return  0;

}
///////////////////////////////////////////

int
Collinear (tPointi a, tPointi b, tPointi c)
{
  return ((int) (AreaSigni( a, b, c ) == 0));
}
/////////////////////////////////////////
// Between
int
Between (tPointi a, tPointi b, tPointi c)
{

  if (!Collinear (a, b, c))
    return (0);

  if (a[XX] != b[XX])
    return ((int)
	    (((a[XX] <= c[XX]) && (c[XX] <= b[XX]))
	     || ((a[XX] >= c[XX]) && (c[XX] >= b[XX]))));
  else
    return ((int)
	    (((a[YY] <= c[YY]) && (c[YY] <= b[YY]))
	     || ((a[YY] >= c[YY]) && (c[YY] >= b[YY]))));
}
/////////////////////////////////////////////
int
Xor (int x, int y)
{
  return ((int) (!x ^ !y));
}
////////////////////////////////

int
IntersectProp (tPointi a, tPointi b, tPointi c, tPointi d)
{
  if (Collinear (a, b, c) || Collinear (a, b, d) || Collinear (c, d, a)
      || Collinear (c, d, b))
    return ( (0));

  return ((int)
	  (Xor (Left (a, b, c), Left (a, b, d))
	   && Xor (Left (c, d, a), Left (c, d, b))));
}
////////////////////////////////////////

int
Intersect (tPointi a, tPointi b, tPointi c, tPointi d)
{
  //int n1,n2,n3,n4;

  if (IntersectProp (a, b, c, d))
    return ( (1));

  else
    {
      if (Between (a, b, c) || Between (a, b, d) || Between (c, d, a)
	  || Between (c, d, b))
	return ( (1));
      else
	return ( (0));
    }
}

////////////////////////////////////////////////////////////

int
Diagonalie (tVertex a, tVertex b, tVertex vertices)
{
  tVertex c, c1;

  // For each edge (c,c1) of P
  c = vertices;

  do
    {
      c1 = c->next;
      // Skip edges incident to a or b
      if ((c != a) && (c1 != a) && (c != b) && (c1 != b)
	  && Intersect (a->v, b->v, c->v, c1->v))
	return ( (0));

      c = c->next;

    }
  while (c != vertices);

  return ( (1));
}

///////////////////////////////////////////////


int
Diagonal (tVertex a, tVertex b, tVertex vertices)
{

  if (InCone (a, b) && InCone (b, a) && Diagonalie (a, b, vertices))
    return(1);
      else
	return(0);

}


//////////////////////////////////////////////////

// Initialize v->ear for all vertices

void
EarInit (tVertex vertices)
{
  tVertex v0, v1, v2;

  v1 = vertices;
  do
    {
      v2 = v1->next;
      v0 = v1->prev;
      v1->ear = Diagonal (v0, v2, vertices);
      v1 = v1->next;
    }
  while (v1 != vertices);
}



///////////////////////////////////////////////////////
// Calcul d'angle entre 3 points reels
// sachant que:
// cos(angle) = ba.bc / ||ba||*||bc||
///////////////////////////////////////////////////////
real
Angle3d (tPointd a, tPointd b, tPointd c)
{
   real Tol = 1.0e-5;		// Pour la comparaison des rÃ©Ã©ls

  real num, den, quot;

  num = ((real) (b[XX] - a[XX]) * (real) (b[XX] - c[XX])) +
    ((real) (b[YY] - a[YY]) * (real) (b[YY] - c[YY]));
  den = sqrt ((real) (a[XX] - b[XX]) * (real) (a[XX] - b[XX]) +
	      (real) (a[YY] - b[YY]) * (real) (a[YY] - b[YY])) *
    sqrt ((real) (c[XX] - b[XX]) * (real) (c[XX] - b[XX]) +
	  (real) (c[YY] - b[YY]) * (real) (c[YY] - b[YY]));
#ifdef DEBUG
  Rprintf ("DEBUG angle3 a=%g %g\n", a[XX], a[YY]);
  Rprintf ("DEBUG b=%g %g\n", b[XX], b[YY]);
 Rprintf ("DEBUG c=%g %g\n", c[XX], c[YY]);
 Rprintf ("DEBUG  num=%30.18f den=%30.18f \n",
	   num, den);

  Rprintf ("fabs %g Tol %g\n", fabs (num - den), Tol)
#endif


  if ((realequal (num, den) == True) ||
      (realequal (den, 0.0) == True))
    {
#ifdef DEBUG
      Rprintf ("Il y a des points confondus num=%30.18f den=%30.18f \n");
#endif

      // Il y a des points confondus
      return (0.0);
    }


  quot = num / den;
#ifdef DEBUG
      Rprintf ( "quot  %30.18f\n", quot);
#endif

  if ( (quot > (1.0+Tol)) ||
       (quot < (-1.0-Tol)))
    {
    REprintf ( "Internal error:  quot=%30.18f\n", quot);
    return(-1);
    }



  if ( quot > (1.0) )
    {
      Rprintf ( "=-1: %30.18f\n", quot);
      quot=+1.0;
    }
  else
    {
      if ( quot < (-1.0))
    {
      Rprintf ( "=+1: %30.18f\n", quot);
      quot=-1.0;
    }
    }

  return (acos (quot));
}				// fin Angle3d

///////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////
// Calcul d'angle entre 3 points entiers
/////////////////////////////////////////////////////
real
Angle3i (tPointi a, tPointi b, tPointi c)
{
  tPointd ad, bd, cd;
  ad[XX] = (real) (a[XX]);
  ad[YY] = (real) (a[YY]);
  bd[XX] = (real) (b[XX]);
  bd[YY] = (real) (b[YY]);
  cd[XX] = (real) (c[XX]);
  cd[YY] = (real) (c[YY]);
  return (Angle3d (ad, bd, cd));
}				// fin Angle3i

////////////////////////////////////////////////////
// Fonction:
// Imprimer la chaine des sommets vertices
///////////////////////////////////////////////////
void
PrintVertices (tVertex vertices)
{
  tVertex v;

  v = vertices;
  do
    {
      Rprintf ("%d\t%ld\t%ld\n", v->vnum, v->v[XX], v->v[YY]);
      v = v->next;
    }
  while (v != vertices);
}

///////////////////////////////////////////////////////
// Fonction:
// Creer la liste chainee (vertices)
// des sommets orientes anti-clockwise du polygone d'indice ipoly.
// Si verbose=T, afficher le min et max des coordonnees.
// Arguments d'entree:
// ipoly: indice du poly a considerer.
// nsom: le nombre de ses sommets.
// Poly: pour chaque polygone, pour le sous-polygone d'indice 0,
// pour chacun de ses sommets, ses coordonnees.
// A cette etape, les sous-polys convexes n'etant pas encore
// determines, seul le sous-polygone d'indice 0 est rempli.
// Les sommets y sont orientes clockwise.
// verbose: si T,  afficher le min et max des coordonnees.
// Arguments de sortie:
// vertices: liste chainee des sommets orientes anti-clockwise.
//////////////////////////////////////////////////////////


void
ReadVertices (int verbose, int nsommets,
	      double *Ad, tVertex vertices)
{
  tVertex v;
  int isom, xminmax[DIM], yminmax[DIM];



  // Initialisation des minmax 
  xminmax[0] = xminmax[1] = (int) Adp(0, XX);
  yminmax[0] =yminmax[1] =  (int)Adp(0, YY);

  // Creation de vertices et mise dans le sens anti-clockwise
  // des coordonnees.
  // Le 1ier sommet est le meme que celui dans Poly[ipoly][0].
  for (isom = 0; isom < nsommets; isom++)
    {
      // MakeNullVertex cree un tVertex et l'insere apres vertices
      // (si vertices n'existe pas encore, on lui affecte
      // le  tVertex cree, les 2 pointeurs prev et next pointant
      // sur lui-meme.) 
      v = MakeNullVertex (vertices);
      v->v[XX] =  (int)Adp(isom, XX);
      v->v[YY] =  (int)Adp(isom, YY);
      if (isom == 0)
	v->vnum = 0;
      else
	v->vnum = nsommets - isom;
      // Calcul dans la foulee, des min-max si verbose.
      if (verbose == 1)
	{
	  xminmax[0] = MIN (v->v[XX], xminmax[0]);
	  xminmax[1] = MAX (v->v[XX], xminmax[1]);
	  yminmax[0] = MIN (v->v[YY], yminmax[0]);
	  yminmax[1] = MAX (v->v[YY], yminmax[1]);
	}			// fin verbose
    }				// fin isom

  if (verbose == 1)
    {
      PrintVertices (vertices);
      Rprintf ("%%Bounding box:\n");
      Rprintf ("xmax = %d; xmin = %d; difference: %d\n",
	      xminmax[1], xminmax[0], xminmax[1] - xminmax[0]);
      Rprintf ("ymax = %d; ymin = %d; difference: %d\n",
	      yminmax[1], yminmax[0], yminmax[1] - yminmax[0]);

    }				// fin verbose


}				// fin ReadVertices
