/*--------------- COPYRIGHT ------------------------
| INRA - Unité MIA de Jouy en Josas                |
--------------------------------------------------*/

///////////////////////////////////////////////////////

#include <stdio.h>
#include <limits.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include <stdlib.h>


////////////////////////////////////////////////////
// Les fonctions de l'API R
////////////////////////////////////////////////////
extern char * REprintf( char *, ...);
extern char * Rprintf(char *, ...);
extern char *R_alloc(size_t n, int size);

////////////////////////////////////////////////////
// Les includes
////////////////////////////////////////////////////
#include "lesdefines.h"
#include "lestypes.h"
#include "lesmacros.h"
#include "geom.h"

//////////////////////////////////////////////////////
#define Adp(a, b) ( Ad[(*nia) * (b) + (a)])
#define Bdp(a, b) ( Bd[(*nib) * (b) + (a)])
//////////////////////////////////////////////////////

typedef enum
  { distinct, included, cross, sur } tRet;

//////////////////////////////////////////////////////
int
InPolyConvex (tPointd t, tPointi * M, int k)
{
  tPointd T1, T2;
  int i;
  i = 0;


  do
    {
      Assd (T1, M[i]);
      Assd (T2, M[i + 1]);
      if (!LeftOnd (T1, T2, t))
	{
	  return ( (int) False);
	}
      i++;
    }
  while (i < (k - 1));

  Assd (T1, M[k - 1]);
  Assd (T2, M[0]);
  if (!LeftOnd (T1, T2, t))
    {
      return ( (int) False);
    }

  return ( (int) True);
}

//////////////////////////////////////////////////////

////////////////////////////////////////////////////
// La fonction maitre
// Renvoie un code selon que le poly Bd est inclus,
// chevauche, contient, est distinct de Ad
////////////////////////////////////////////////////
void
inconvex ( int * MAX_VERTICES, 
	   double *Ad, double *Bd,
	   int *nia, int *nib,
	   int *inclus)
{
  int i,j, nsom;
  tRet  retour= distinct;

  tPointi *  tPol;
  tPointd pPoint;

 // Allocations:
  tPol = (tPointi *)
    R_alloc((*MAX_VERTICES), sizeof(tPointi));

  *inclus= retour;

  if ((*nia > (*MAX_VERTICES)) ||
      (*nib > (*MAX_VERTICES)))
    {
      REprintf ("Polygons have too many vertices\n");
      *inclus =-1;
      return;
    }

  // Mettre Ad dans un tPolygoni
  for (i =0; i < *nia; i++)
    {
    for (j=0 ; j< DIM; j++)
        tPol[i][j] = (int) Adp(i, j);

    }


  nsom=0;
  // Boucle sur les sommets de B
  for (i =0; i < *nib; i++)
  {
    for (j=0 ; j< DIM; j++)
      pPoint[j] = Bdp(i, j);


    if (InPolyConvex (pPoint, tPol, *nia) == 1)
       nsom++;
  } // fin Boucle sur les sommets de B

  if (nsom == *nib)
      retour = included;
else
  if (nsom >0)
      retour = cross;

  else{
    // B n'est pas inclus dans A et aucun de ses sommets est dans A

  // Voir si des sommets de A sont dans B
  // Mettre Bd dans un tPolygoni
  for (i =0; i < *nib; i++)
    {
    for (j=0 ; j< DIM; j++)
        tPol[i][j] = (int) Bdp(i, j);

    }


  nsom=0;
  // Boucle sur les sommets de A
  for (i =0; i < *nia; i++)
  {
    for (j=0 ; j< DIM; j++)
      pPoint[j] = Adp(i, j);
    if (InPolyConvex (pPoint, tPol, *nib) == 1)
       nsom++;
  } // fin Boucle sur les sommets de A

  if (nsom == *nia)
      retour = sur;
else
    if (nsom >0)
      retour = cross;

  }

      *inclus= (int)retour;


} // fin inconvex
